<?php
namespace Zeedhi\Tools\DatabaseReader\Adapters;

use Doctrine\DBAL\Connection;
use Zeedhi\Tools\DatabaseReader\AdapterReader;
use Zeedhi\Tools\DatabaseReader\DTO;
use Zeedhi\Tools\DatabaseReader\Exception;

class Oracle extends AdapterReader{

    protected static $deleteRuleConversion = array(
        'CASCADE'   => DTO\Constraints\ForeignKey::DELETE_RULE_CASCADE,
        'NO ACTION' => DTO\Constraints\ForeignKey::DELETE_RULE_NO_ACTION,
        'RESTRICT'  => DTO\Constraints\ForeignKey::DELETE_RULE_RESTRICT,
        'SET NULL'  => DTO\Constraints\ForeignKey::DELETE_RULE_SET_NULL
    );

    const CONSTRAINT_TYPE_PK = "P";
    const CONSTRAINT_TYPE_FK = "R";
    const CONSTRAINT_TYPE_UNIQUE = "U";

    protected static $columnDataTypeConversion = array(
        'BLOB'      => DTO\Column::DATA_TYPE_BINARY,
        'CLOB'      => DTO\Column::DATA_TYPE_CLOB,
        'NUMBER'    => DTO\Column::DATA_TYPE_INTEGER,
        'VARCHAR2'  => DTO\Column::DATA_TYPE_STRING,
        'CHAR'      => DTO\Column::DATA_TYPE_STRING_FIXED,
        'DATE'      => DTO\Column::DATA_TYPE_DATETIME,
        'FLOAT'     => DTO\Column::DATA_TYPE_FLOAT,
        'RAW'       => DTO\Column::DATA_TYPE_BINARY,
        'LONG RAW'  => DTO\Column::DATA_TYPE_BINARY,
        // There's no datetime column type in oracle.
        //'DATETIME'  => DTO\Column::DATA_TYPE_DATETIME,
        //DTO\Column::DATA_TYPE_BOOLEAN, // Not 'inferable' type.
    );

    /** @var \Doctrine\DBAL\Connection */
    protected $connection;
    /** @var OracleSequenceNameProvider */
    protected $sequenceNameProvider;
    /** @var string */
    protected $ownerName;

    public function __construct(Connection $connection, $sequenceNameProvider = null) {
        $this->connection = $connection;
        $this->sequenceNameProvider = $sequenceNameProvider ?: new OracleSequenceNameProvider();
        $this->ownerName = $this->connection->getDatabase();
    }

    /**
     * @param string $ownerName
     */
    public function setOwnerName($ownerName) {
        $this->ownerName = $ownerName;
    }

    /**
     * Get name from all tables in database.
     *
     * @return string[] With name of all tables.
     */
    public function getAllTableNames() {
        $tableRows = $this->connection->fetchAll(
            "SELECT TABLE_NAME FROM ALL_TABLES WHERE OWNER = :OWNER",
            array('OWNER' => $this->ownerName)
        );
        $tableNames = array();
        foreach($tableRows as $dbTable){
            $tableNames[] = $dbTable['TABLE_NAME'];
        }

        return $tableNames;
    }

    /**
     * Get a table DTObject with all data about chosen table.
     *
     * @param string $tableName Name from chosen table.
     *
     * @return DTO\Table
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception When table was not found in database.
     */
    public function getTable($tableName) {
        $tableRow = $this->getTableRow($tableName);
        $DTOTable = new DTO\Table();
        $DTOTable->setName($tableRow['TABLE_NAME']);
        return $DTOTable;
    }

    /**
     * Return constraints rows of given table that belongs to chosen table.
     *
     * @param string $tableName      Name of table that belongs constraints.
     * @param string $constraintType Type of constraint, must be on of self::CONSTRAINT_TYPE_*.
     *
     * @return array of Rows that are array too.
     */
    protected function getConstraintRow($tableName, $constraintType) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT CONSTRAINT_NAME, R_CONSTRAINT_NAME, DELETE_RULE FROM
            ALL_CONSTRAINTS WHERE CONSTRAINT_TYPE = :CONSTRAINT_TYPE AND TABLE_NAME = :TABLE_NAME AND OWNER = :OWNER");
        }

        $statement->execute(array(
            "TABLE_NAME" => $tableName,
            "CONSTRAINT_TYPE" => $constraintType,
            "OWNER" => $this->ownerName
        ));
        return $statement->fetchAll();
    }

    protected function getPrimaryKeyRow($tableName) {
        $primaryKeyRow = $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_PK);
        if (!$primaryKeyRow) throw Exception::tableWithoutPrimaryKey($tableName);
        return reset($primaryKeyRow);
    }

    protected function getForeignKeyRows($tableName) {
        return $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_FK);
    }

    protected function getUniqueConstraintRows($tableName) {
        return $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_UNIQUE);
    }

    protected function getDTOForeignKeyFromRow($foreignKeyRow){
        if(!isset(self::$deleteRuleConversion[$foreignKeyRow['DELETE_RULE']])) {
            throw Exception::unsupportedDeleteRule($foreignKeyRow['DELETE_RULE']);
        }
        $deleteRule = self::$deleteRuleConversion[$foreignKeyRow['DELETE_RULE']];
        $DTOForeignKey = new DTO\Constraints\ForeignKey(
            $foreignKeyRow['CONSTRAINT_NAME'],
            $foreignKeyRow['R_CONSTRAINT_NAME'],
            $deleteRule
        );
        return $DTOForeignKey;
    }

    protected function getDTOUniqueConstraintsFromRows($uniqueConstraintRows) {
        $DTOUniqueConstraint = array();
        foreach($uniqueConstraintRows as $uniqueConstraintRow) {
            $DTOUniqueConstraint[] = new DTO\Constraints\Unique($uniqueConstraintRow['CONSTRAINT_NAME']);
        }

        return $DTOUniqueConstraint;
    }

    /**
     * @param $tableName
     *
     * @return mixed
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception
     */
    protected function getTableRow($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT TABLE_NAME FROM ALL_TABLES WHERE OWNER = :OWNER AND TABLE_NAME = :TABLE_NAME");
        }
        $statement->execute(array('TABLE_NAME' => $tableName, 'OWNER' => $this->ownerName));
        $tableRow = $statement->fetch(\PDO::FETCH_ASSOC);
        if (!$tableRow) throw Exception::tableNotFound($tableName);
        return $tableRow;
    }

    /**
     * @param $tableName
     * @return array
     */
    protected function getColumnRows($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT COLUMN_NAME, COLUMN_ID, DATA_TYPE, DATA_LENGTH,
                                                     DATA_PRECISION, DATA_SCALE, NULLABLE, DATA_DEFAULT
                                                     FROM ALL_TAB_COLUMNS
                                                     WHERE TABLE_NAME = :TABLE_NAME
                                                     AND OWNER = :OWNER");
        }

        $statement->execute(array('TABLE_NAME' => $tableName, 'OWNER' => $this->ownerName));
        $columnsRow = $statement->fetchAll();
        return $columnsRow;
    }

    /**
     * Factory a DTO\Column from a row of Oracle Dictionary.
     *
     * @param array $columnRow With positions COLUMN_NAME, DATA_DEFAULT, DATA_LENGTH, DATA_PRECISION, DATA_SCALE, DATA_TYPE, NULLABLE
     *
     * @return DTO\Column
     */
    protected function getDTOColumnFromRow($columnRow) {
        $DTOColumn = new DTO\Column();
        $DTOColumn->setName($columnRow['COLUMN_NAME']);
        $DTOColumn->setDataType(self::$columnDataTypeConversion[$columnRow['DATA_TYPE']]);
        $DTOColumn->setDataDefault($columnRow['DATA_DEFAULT']);
        $DTOColumn->setDataLength($columnRow['DATA_LENGTH']);
        $DTOColumn->setDataPrecision($columnRow['DATA_PRECISION']);
        $DTOColumn->setDataScale($columnRow['DATA_SCALE']);
        $DTOColumn->setNullAble($columnRow['NULLABLE']);
        return $DTOColumn;
    }

    /**
     * @param $primaryKeyRow
     * @return DTO\Constraints\PrimaryKey
     */
    protected function getDTOPrimaryKeyFromRow($primaryKeyRow) {
        return new DTO\Constraints\PrimaryKey($primaryKeyRow['CONSTRAINT_NAME']);
    }

    protected function getDTOConsColumnFromRow($consColumnRow) {
        $DTOConsColumn = new DTO\Constraints\ConsColumn($consColumnRow['COLUMN_NAME']);
        return $DTOConsColumn;
    }

    protected function getConsColumnsRow($constraintName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT COLUMN_NAME, POSITION FROM ALL_CONS_COLUMNS
                                                    WHERE CONSTRAINT_NAME = :CONSTRAINT_NAME
                                                     AND OWNER = :OWNER");
        }

        $statement->execute(array(":CONSTRAINT_NAME" => $constraintName, 'OWNER' => $this->ownerName));
        return $statement->fetchAll();
    }

    /**
     * Get foreign key from chosen table with given name.
     *
     * @param string $tableName Name from chosen table.
     * @param string $constraintName Name of constraint.
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception When constraint was not found in database.
     *
     * @return DTO\Constraints\ForeignKey
     */
    public function getForeignKeyByNameAndTableName($constraintName, $tableName) {
        $statement = $this->connection->prepare("SELECT CONSTRAINT_NAME, R_CONSTRAINT_NAME, DELETE_RULE FROM ALL_CONSTRAINTS WHERE TABLE_NAME = :TABLE_NAME AND CONSTRAINT_NAME = :CONSTRAINT_NAME AND OWNER = :OWNER");
        $statement->execute(array(':TABLE_NAME' => $tableName, ':CONSTRAINT_NAME' => $constraintName, 'OWNER' => $this->ownerName));
        $constraintRow = $statement->fetch(\PDO::FETCH_ASSOC);
        if (!$constraintRow) throw Exception::constraintNotFoundInDatabase($tableName, $constraintName);
        $DTOForeignKey = $this->getDTOForeignKeyFromRow($constraintRow);
        $this->completeConsColumnForConstraint($DTOForeignKey);
        return $DTOForeignKey;
    }

    /**
     * Get the sequence of that table.
     *
     * @param string $tableName Name of table.
     * @param string $columnName Name of column that uses that sequence.
     *
     * @throws Exception When a sequence doesn't exist.
     *
     * @return DTO\Sequence
     */
    public function getSequenceFromTableAndColumn($tableName, $columnName) {
        try {
            $sequenceName = $this->getSequenceName($tableName, $columnName);
            $statement = $this->connection->prepare("SELECT SEQUENCE_NAME, MIN_VALUE, MAX_VALUE, INCREMENT_BY FROM
                                                    DBA_SEQUENCES WHERE SEQUENCE_NAME = :SEQUENCE_NAME AND SEQUENCE_OWNER = :OWNER");
            $statement->execute(array(':SEQUENCE_NAME' => $sequenceName, 'OWNER' => $this->ownerName));
            $sequence = $statement->fetch(\PDO::FETCH_ASSOC);
            if(!$sequence){
                throw Exception::sequenceDoesNotExist($tableName, $columnName);
            }
        } catch (\Exception $e) {
            return NULL;
        }

        return new DTO\Sequence($sequence["SEQUENCE_NAME"],$sequence["MIN_VALUE"],$sequence["INCREMENT_BY"],$sequence["MAX_VALUE"]);
    }

    /**
     * @param $tableName
     * @param $columnName
     * @return string
     */
    protected function getSequenceName($tableName, $columnName) {
        return $this->sequenceNameProvider->getSequenceName($tableName, $columnName);
    }

    /**
     * @return OracleSequenceNameProvider
     */
    public function getSequenceNameProvider() {
        return $this->sequenceNameProvider;
    }

    /**
     * @param OracleSequenceNameProvider $sequenceNameProvider
     */
    public function setSequenceNameProvider($sequenceNameProvider) {
        $this->sequenceNameProvider = $sequenceNameProvider;
    }

    public function getCheckConstraints($tableName) {
        $query = "SELECT CONSTRAINT_NAME, SEARCH_CONDITION FROM ALL_CONSTRAINTS WHERE TABLE_NAME LIKE :TABLE_NAME AND OWNER LIKE :OWNER AND CONSTRAINT_TYPE = 'C'";
        $stmt = $this->connection->prepare($query);
        $stmt->execute(array(':TABLE_NAME' => $tableName, ':OWNER' => $this->ownerName));
        $checkConstraintRows = array_filter($stmt->fetchAll(), function($row) {
            return !preg_match('/"[A-Za-z_]+" IS NOT NULL/', $row['SEARCH_CONDITION']);
        });

        return array_map(function($checkRow) {
            return new DTO\Constraints\Check($checkRow['CONSTRAINT_NAME'],  $checkRow['SEARCH_CONDITION']);
        }, $checkConstraintRows);
    }

    public function getDataFromRefConstraint($constraintName) {
        $stmt = $this->connection->prepare(
            "SELECT
                B.COLUMN_NAME,
                A.CONSTRAINT_NAME,
                A.TABLE_NAME,
                A.CONSTRAINT_TYPE
            FROM
                ALL_CONSTRAINTS A
            JOIN
                ALL_CONS_COLUMNS B
            ON
                A.CONSTRAINT_NAME = B.CONSTRAINT_NAME
            WHERE
                A.CONSTRAINT_NAME = :CONSTRAINT_NAME
            AND
                A.OWNER = :OWNER
            AND
                B.OWNER = :OWNER"
        );
        $stmt->execute(array(':CONSTRAINT_NAME' => $constraintName, ':OWNER' => $this->ownerName));
        return $stmt->fetchAll();
    }
}